function export_vtk(Nodes,Elems,filename,cvals)
  %% exports the geometric data given in Nodes and Elems to a vtk file
  %% if cvals is presented, it is assumed that cevals is the value of
  %% whatever you want to display at the given nodes
  %%
  %%  input:
  %%     Nodes   matrix number_nodes times 3
  %%             contains the coordinates
  %%             of each vertex of the mesh, as no nodenumber is given
  %%             it is assumed all nodes are ordered as they are used
  %%             in the Elems list
  %%     Elems   matrix number_elems times 4
  %%             contains the number of the vertices for each element,
  %%             needs to be consistent with Nodes. If the vertexlist
  %%             does not start with 0, which is needed for paraview,
  %%             the number of the lowest vertex is subtracted from
  %%             the list
  %%             as we have triangles and quadrilaterals we need 4
  %%             entries per element, if a triangle is given in a
  %%             mixed mesh the last entry is set to -1
  %%    filename Name of the output file
  %%    cvals    vector with lengtzh number_nodes 
  %%             contains the values of the whateveryouwanttodisplay
  %%             for each Node
  %% Copyright (C) 2022 W. Kreuzer <wolfgang.kreuzer@oeaw.ac.at>.
  %%
  %%
  %% This program is free software: you can redistribute it and/or modify
  %% it under the terms of the GNU General Public License as published by
  %% the Free Software Foundation, either version 3 of the License, or
  %% (at your option) any later version.
  %
  %% This program is distributed in the hope that it will be useful,
  %% but WITHOUT ANY WARRANTY; without even the implied warranty of
  %% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  %% GNU General Public License for more details.
  %%
  %% You should have received a copy of the GNU General Public License
  %% along with this program.  If not, see <http://www.gnu.org/licenses/>.
  

    if(nargin<4)
      cvals = [];
    end


%    if(reduced)
%      Nodes=Nodes';
%    end

    if( rows(Nodes) < columns(Nodes) )
      % probably wrong format
      Nodes = Nodes';
    end

    
    %% paraviews nodes start with 0, this should fix that
    minE = min(Elems(find(Elems(:)>-1)));

    Elems = Elems - minE;
    
 
    fid = fopen(filename, 'wt');
    fprintf(fid, "# vtk DataFile Version 3.0\n");
    fprintf(fid, "Comment goes here\n");
    fprintf(fid, "ASCII\n");
    fprintf(fid, "\n");
    fprintf(fid, "DATASET POLYDATA\n");

    [n1,n2,n3]=size(Nodes);
    if( n2 == 3 )
      Nodes = Nodes';
      [n1,n2,n3]=size(Nodes);
    end
    fprintf(fid, "POINTS   %d double\n", n3*n2);

    for i=1:n3, 
      for j=1:n2,
        fprintf(fid,"%f %f %f\n",Nodes(1,j,i),Nodes(2,j,i),Nodes(3,j,i));
      end
    end

    % see if there are possibly triangles in E that are marked by -1
    trielems = 0;
    [e1,e2,e3]=size(Elems);

		     % it is assumed that the nodenumber starts with 0
		     % in nodes
   
    for i = 1:e3,   % maybe we have different faces
    		    % do the triangles first
      % there could be some hidden triangles in there where the last
      % index is -1, remember the elem number is alread removed

     

      if(columns(Elems(:,:,i)) == 4) 
	trielems = find(Elems(:,4,i) < 0);
      else
	trielems = [1:rows(Elems(:,:,i))];
      end

      if(isempty(trielems))
	trielems = 0;
	Quads = Elems(:,:,i);
	quadelems = rows(Elems);
      else
	Quads = Elems(:,:,i);
	Quads(trielems,:) = [];
	Etri = Elems(trielems,:,i);
	trielems = rows(Etri);
	if(isempty(Quads))
	  quadelems = 0;
	else
	  quadelems = rows(Quads);
	end
	if(columns(Etri) == 4)
				% get rid of the -1
	  Etri(:,4) = [];
	end

      end

      fprintf(fid, "POLYGONS %d %d\n", trielems + quadelems, ...
	      4*trielems + 5*quadelems);
      

      for j = 1:trielems,
	fprintf(fid,'3 %d %d %d \n', Etri(j,1,i),Etri(j,2,i),Etri(j,3,i));
      end
      				%     keyboard
      

      
      
      if(!isempty(Quads))
%	fprintf(fid, "POLYGONS %d %d\n", rows(Quads), 5*rows(Quads));
	for j = 1:rows(Quads),
	  fprintf(fid,'4 %d %d %d %d\n', Quads(j,1,i),Quads(j,2,i),Quads(j,3,i),Quads(j,4,i));
	end
      end	
    end

    
  
    if(!isempty(cvals))
      % we switched the dimension of the matrices somewhere in the code
      if( length(cvals) == rows( Nodes' ) )
	%% we are good to go
	fprintf(fid, "POINT_DATA %d\n", length(cvals));
	text1 = sprintf('SCALARS dataset1 double\n');
        fprintf(fid, text1);
        fprintf(fid, 'LOOKUP_TABLE default\n');	
	fprintf(fid, '%e\n', cvals');
      end
    end
    fclose(fid)
end
