

function [nelems,avedge] = sphereexample(r,n,f,x0,bchard,patchsphere,rwfact)
  %% 1) creates a sphere with n subdivis1ions  with sound hard or
  %%    sound soft boundary conditions.
  %%    Also the script creates an evaluation mesh around the
  %%    sphere, and a vtk file usable by paraview, so both, mesh and
  %%    evalgrid can be displayed
  %%    Note: For sound soft boundary conditions, the particle
  %%    velocity is displayed on the sphere, but the sound pressure is
  %%    displayed on the evaluation grid
  %%
  %% For constructing the mesh, the  face{*} struct for the different nodes
  %% and elements are used, also a param struct for the parameters of mesh2hrtf
  %%
  %% input:       r  ... radius of the scattering sphere
  %%              n  ... number of subdivision used to create the mesh
  %%              f  ... frequency, may be a vector containing
  %%                     multiple frequencies
  %%             x0 ... possible direction of the plane wave
  %%                    (not used yet, because you can
  %%                    rotate the sphere if you want to change
  %%                    direction), just write 0
  %%         bchard ... boundary condition acoustically hard = 1
  %%                                   acoustically soft = all other
  %%                                   values
  %%    patchsphere ...  originally a sphere based on an icosaeder was used
  %%                     but this has the drawback that lesser
  %%                     subdivisions can be made. Also, as the
  %%                     icosaeder sphere is based on an external tool
  %%                     box, it is not bundled with this example. There is 
  %%                     a different method to create the
  %%                     sphere. In this case a sphere with triangular
  %%                     elements is created by subdividing a cube and
  %%                     projecting it onto the sphere, this has the
  %%                     advantage, that the number of elements is not
  %%                     as restricted as in the case for an
  %%                     icosphere. Warning: If one would want to use
  %%                     quadrilaterals here, the elements at the
  %%                     edges of the projected cube, will not lie in
  %%                     a plane. However patchsphere = 1 uses this
  %%                     sphere
  %%         rwfact ... factor to determine the accuracy of the FMM, 5
  %%                    would approximately mean float, 10 double, default
  %%                    is 1.8, which is a good compromise between
  %%                    accuracy and speed
  %%
  %%  output: nelems ...  number of elements used for discretizing the
  %%                      sphere. Was used for the order plot
  %%          avedge ...  average edge length
  %%
  %%  Oh yes, and another thing, the script creates the files
  %%  ouputBEM_%d.vtk for each frequency and outputEval_%d.vtk to
  %%  display the results
  %%
  %%  the figures in this tar ball were made with
  %%  sphereexample(1,25,[200,1000],0,0,1) for the soft sphere
  %%  and
  %%  sphereexample(1,25,[200,1000],0,1,1) for the hard sphere
  %%  the default setting (no input) should also produce the hard
  %%  sphere
  %%
  %%
  %% Copyright (C) 2022 W. Kreuzer <wolfgang.kreuzer@oeaw.ac.at>.
  %%
  %%
  %% This program is free software: you can redistribute it and/or modify
  %% it under the terms of the GNU General Public License as published by
  %% the Free Software Foundation, either version 3 of the License, or
  %% (at your option) any later version.
  %
  %% This program is distributed in the hope that it will be useful,
  %% but WITHOUT ANY WARRANTY; without even the implied warranty of
  %% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  %% GNU General Public License for more details.
  %%
  %% You should have received a copy of the GNU General Public License
  %% along with this program.  If not, see <http://www.gnu.org/licenses/>.

  if(nargin < 7)
    rwfact = 1.8;
    if(nargin < 6)
      patchsphere = 0;
      if(nargin < 5)
	bchard = 1;
	if(nargin < 4)
	  x0 = [0,0,1.5];
	  if( nargin < 3 )
	    f = [200,1000];
	    if(nargin < 2)
	      n = 25;   % about n = 4 is about 5000 elements for
		       % icosphere. for the patchsphere 15 is about
		       % 2500 elements, I guess
	      if(nargin < 1)
		r = 1.0
	      end
	    end
	  end
	end
      end
    end
  end

  printstuff = 0;  % don't create tables and stuff
  writestuff = 0;
  close all
  if(r == [])
    r = 1.0
  end

  rwfact

  if(n < 0)
    n = 0
  end
  nfreqsteps = length(f);
  
  %% set the frequency
  
  rho = 1.3;
  c = 340;
  
  removedbl = 0;
  
  %% create the nodes and elements for the sphere
  %% Nodenumbers will start with 1
  %% take one of the next two possibibities
  %% a) Icosphere is the classical open, however there is a
  %% restriction on the number of elements
  %% b) that a projection of a cube to the sphere, works for
  %% trangles
  %[NSphere,ESphere] = icosphere(n);
  [NSphere,ESphere,~,~,avedge] = create_patchsphere(n);


  
  nelems = rows(ESphere);
  # export the mesh to vtk so it can be desplayed using paraview
  export_vtk(NSphere,ESphere,"Sphere.vtk");


  
  %% calculate the collocation nodes
  %% there is a slight problem, that in principle the collocnodes
  %% are not on the unit sphere, but the error can be neglected
  cnodes = 1/3.0 * (NSphere(ESphere(:,1),:) + NSphere(ESphere(:,2),:) + NSphere(ESphere(:,3),:));

  % lets create the faces, but let the bem nodes be the first face,
  % since the nodes will not be ordered later, the first nnodes are
  % always the nodes of the sphere, helps when displaying data
  
  face{1}.nodes = NSphere';
  %face{1}.elem = [ESphere,-1*ones(rows(ESphere),1)];
  face{1}.elem = ESphere;

  face{1}.type = 0;
  face{1}.group = 0;
  if( bchard ) % soundhard
    face{1}.bctype = 'VELO';
    face{1}.bcval = 0.0;
  else % soundsoft
    face{1}.bctype = 'PRES';
    face{1}.bcval = 0.0;
  end
  face{1}.bccurve = [-1,-1];
  

  %% create the evalnodes, remember in general NC.inp
  %% needs evalelements but the elements themselves are never
  %% used, except for defining the nodetype and for plotting
  %% lets not care about elements right now, just group the nodes into
  %% elements

  %% just to make creation of evalelements easier
  %neval = floor(neval/3) * 3;

  % you want random you get random nodes, just uncomment the next sections
  %   Nphi = rand(neval,1)* 2 * pi;
  %   Ntheta = rand(neval,1) * pi;
  %   Nr = rand(eval,1) * (r2 - r1) + 1.1*r1;
  %   NE(:,1) = Nr .* sin(Ntheta) .* cos(Nphi);
  %   NE(:,2) = Nr .* sin(Ntheta) .* sin(Nphi);
  %   NE(:,3) = Nr .* sin(Ntheta);
  %face{2}.nodes = NE';
  %EE = reshape([1:neval],3,neval/3);
  %face{2].elem = EE';

  %%  make the evalgrid using quadrilaterals
  
  triangles = 0;  % evalelements should be quadrilatereals
  		  % make something nice to plot, construct the
		  % evaluation grid using 4 parts that will be subdivided
  
  n1 = 60*3+1;
  n2 = 17*3;
  x1 = [-3,0,1.2];
  x2 = [3,0,1.2];
  x3 = [3,0,3];
  x4 = [-3,0,3];
  x = [x1',x2',x3',x4'];

  [face{2}.nodes,face{2}.elem] = fillface(x,n1,n2+1,triangles);

  x1 = [-3,0,-1.2];
  x2 = [3,0,-1.2];
  x3 = [3,0,-3];
  x4 = [-3,0,-3];
  x = [x1',x2',x3',x4'];  % we don't care about the direction, since
			  % they are only eval elements and no normal
			  % vector is needed
  [face{3}.nodes,face{3}.elem] = fillface(x,n1,n2+1,triangles);

  n1 = 18*3;n2 = 18*3+1;
  x1 = [-3,0,-1.2];
  x2 = [-1.2,0,-1.2];
  x3 = [-1.2,0,1.2];
  x4 = [-3.0,0,1.2];
  x = [x1',x2',x3',x4'];  % we don't care about the direction
  [face{4}.nodes,face{4}.elem] = fillface(x,n2,n1+1,triangles);

  n1 = 18*3;n2 = 18*3+1;
  x1 = [1.2,0,-1.2];
  x2 = [3.0,0,-1.2];
  x3 = [3.0,0,1.2];
  x4 = [1.2,0,1.2];
  x = [x1',x2',x3',x4'];  % we don't care about the direction
  [face{5}.nodes,face{5}.elem] = fillface(x,n1,n2+1,triangles);

  
  
  
  %% as already mentioned, we don't really care about the
  %% eval elements, use the default bc, just do that so octave does
  %% not complain about having some bad definition
  for i = 2:5,
    face{i}.type = 2;
    face{i}.group = 1;
    face{i}.bctype = "VELO";
    face{i}.bcval = 0.0;
  end


  % merge the different faces, if removedbl = 0 double nodes are not
  % removed, which is okay for evalelements. The problem with
  % removedble = 1 is the fact, that a ordering of the nodes is used
  % for that, and we would loose the nice property that the BEM nodes
  % come first. If you are not satisfied with this solution, you
  % could use different files for bem and eval, remember NumCalc
  % allows that
  [N,E] = remove_patches(face,removedbl);


   %% be careful E has in the last column the facenumber
  
  export_vtk(N,E,"SphereEval.vtk"); % for paraview



  
  %% we haven't set any frequency,
  %%   f = 200Hz
  %%   c = 340;
  %% rho = 1.3;
  
  % set the frequency parameter
  
  param.freq0 = 0.0
  param.freqstep = 1;  % why not
  param.n_freq = length(f);
  param.freqcrv = [];
  for i = 1:length(f),
    param.freqcrv = [param.freqcrv;i f(i) 0.0];
  end
  % write all the info the NC.inp
  make_inpfile_nc(face,param,[],0,1);

 
  
  k = 2 * pi * f / 340;
 
  %% calculate the solution
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %% Here is the only thing that you need to do
  %% enter the right PATH where you have installed NumCalc
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  mystring = sprintf("~/Mesh2HRTF/current/mesh2hrtf/NumCalc/Source/NumCalc");


  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %%
  %% start numcalc
  %%
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  system(mystring);

  %% okay here's the thing, we took a slight shortcut by not removing
  %% the double nodes in the evalgrid.
  %% this has the advantage that the nodes have not been ordered with
  %% respect to their x-coordinate and the first rows(NSphere) Nodes
  %% are on the sphere

  %% if you use matlab, you need size here
  nbem = rows(NSphere);
  ebem = rows(ESphere);

 
  evalnodes = N(nbem + 1:end,:);
  evalelems = E(ebem+1:end,:);
  bemnodes = N(1:nbem,:);
  bemelems = E(1:ebem,:);
  %meval = min(min(evalelems));
  %evalelems = evalelems - meval + 1;
		 
  for fstep = 1:nfreqsteps,
    
    if(bchard)
      filename = sprintf("be.out/be.%d/pBoundary",fstep);
      pcolloc = importdata(filename);
		 %      pcolloc = importdata("be.out/be.1/pBoundary");
      pcolloc = pcolloc.data;
      pcolloc(1:2,:) = [];
      pcolloc = (pcolloc(:,2) + 1.0i * pcolloc(:,3));
 				% pressure to dB
      pcolloc = 20 * log10 (  abs(pcolloc)/2.0e-5 );
 
    else
      filename = sprintf("be.out/be.%d/vBoundary",fstep);
		   %    vcolloc = importdata("be.out/be.1/vBoundary");
      vcolloc = importdata(filename);
      vcolloc = vcolloc.data;
      vcolloc(1:2,:) = [];
      vcolloc = (vcolloc(:,2) + 1.0i * vcolloc(:,3))
      filename = sprintf("VeloBEM_%d.vtk",fstep);
      % write the particle velocity normal to the surface
      %vcolloc = (vcolloc(:,2) + 1.0i * vcolloc(:,3)) * 2.0i * pi * f * rho;
    end

    % slight problem, we have the value at the collocation nodes, we
    % would need them on the BEM nodes, thus, interpolation
    % this may not be optimal matlab code, but it does the job
    for nodenumber = 1:length(bemnodes),
      % now here's the trick, since we know how bemelems was created
      % it contains no jumps and node number start with 1
      [i,j] = find( bemelems == nodenumber );
	 % i should now contain all elements containing the nodenumber
	 % just take the mean value
      if(bchard)
	pBEM(nodenumber) = mean(pcolloc(i));
      else
	%% okay this is very naughty, because the notation is rubbish
	%% but to be honest, I didn't want to think to much about the
	%% correct notation for sound soft spheres, pBEM is written,
	%% so it is called this way, even if the velocity is displayed
	pBEM(nodenumber) = mean(vcolloc(i));
      end
    end
    
    %% b) value at the evalnodes
    filename = sprintf("be.out/be.%d/pEvalGrid",fstep);
    ap = importdata(filename);
    ap = ap.data;
    ap(1:2,:) = [];
    peval = ap(:,2) + 1.0i * ap(:,3);
    %% pressure -> dB
    peval = 20 * log10 ( peval/2.0e-5 );
    %% now export everything to vtk
    filename1 = sprintf("outputBEM_%d.vtk",fstep);
    filename2 = sprintf("outputEval_%d.vtk",fstep);

    export_vtk(bemnodes,bemelems,filename1,pBEM)
    export_vtk(evalnodes, evalelems,filename2,peval)
  
  end % end frequency loop

end

function [N,E,mined,maxed,aved] = create_patchsphere(n)
  %% creates a sphere as a projetion from a cube,
  %% we assume that the sphere has radius 1, for another radius the
  %% nodes have to be scaled
  %%
  %% input   n: integer, number of subdivisions of each face of the cube
  %%
  %% ouput   N: nodes of the sphere
  %%         E: Elements of the sphere (containing the nodenumbers as
  %%            vertices), be careful these elements need to be
  %%            triangles, quadrilaterals will not be in one plane
  %%         mined, maxed, aved : minimum, maximum, average edge length
  % create the original cube, if the radius is one, the length of the
  % cube edges should be 2/sqrt(3)

  a = 2/sqrt(3);
  a2 = 1/sqrt(3);

  % vertices of the first face for the cube
  x1 = [-a2,a2,-a2];
  x2 = [a2 ,a2,-a2];
  x3 = [a2 ,a2,a2];
  x4 = [-a2,a2,a2];
  x = [x1',x4',x3',x2'];
  [face{1}.nodes,face{1}.elem] = fillface(x,n,n,1);
  % second face
  x1 = [a2,a2,-a2];
  x2 = [a2,-a2,-a2];
  x3 = [a2,-a2,a2];
  x4 = [a2,a2,a2];
  x = [x1',x4',x3',x2'];
  [face{2}.nodes,face{2}.elem] = fillface(x,n,n,1);

  x1 = [a2,-a2,-a2];
  x2 = [-a2,-a2,-a2];
  x3 = [-a2,-a2,a2];
  x4 = [a2,-a2,a2];
  x = [x1',x4',x3',x2'];
  [face{3}.nodes,face{3}.elem] = fillface(x,n,n,1);

  x1 = [-a2,-a2,-a2];
  x2 = [-a2,a2,-a2];
  x3 = [-a2,a2,a2];
  x4 = [-a2,-a2,a2];
  x = [x1',x4',x3',x2'];
  [face{4}.nodes,face{4}.elem] = fillface(x,n,n,1);

  x1 = [-a2,a2,a2];
  x2 = [a2,a2,a2];
  x3 = [a2,-a2,a2];
  x4 = [-a2,-a2,a2];
  x = [x1',x4',x3',x2'];
  [face{5}.nodes,face{5}.elem] = fillface(x,n,n,1);

  x1 = [a2,a2,-a2];
  x2 = [-a2,a2,-a2];
  x3 = [-a2,-a2,-a2];
  x4 = [a2,-a2,-a2];
  x = [x1',x4',x3',x2'];
  [face{6}.nodes,face{6}.elem] = fillface(x,n,n,1);

 
  % merge the 6 faces and get rid of double nodes
  [N,E] = remove_patches(face);


  %export_vtk(N,E,"Cube.vtk");

  %project everything to the sphere
  
  if(columns(N) == 3)
    nN = norm(N,'rows');
    N = N./nN;
  end

  export_vtk(N,E,"Sphere.vtk");

  mined = 1e10;
  maxed = 0;
  aved = 0;
  %% get an estimate for the edgelengths
  evec = norm( N( E(:,1),: ) - N( E(:,2),: ),'rows');
  mined = min( mined, min( evec ) );
  maxed = max( maxed, max( evec ) );
  aved = sum( evec );

  evec = norm( N( E(:,1),: ) - N( E(:,3),: ),'rows');
  mined = min( mined, min( evec ) );
  maxed = max( maxed, max( evec ) );
  aved = (aved + sum(evec));
  
  evec = norm( N( E(:,2),: ) - N( E(:,3),: ),'rows');
  mined = min( mined, min( evec ) );
  maxed = max( maxed, max( evec ) );
  aved = (aved + sum(evec))/rows(E)/3;
end


function [nodes,elem] = fillface(x,N1,N2,triangles)
% function to fill face{i} with elements and 
% x          ... coords of the vertices
%                  Matrix 3 x 4 containing the
%                  coordinates of the patch vertices
% N1        ... number of subdivisions in the first direction
% N2        ... number of subdivisions in the second direction
% triangles ... fill the face with triangles instead of rectangles
    
    if(nargin < 4)
        triangles = 0;
    end
    ncounter = 1;
    ecounter = 1;

    gx = linspace(0,1,N1); 
    nx = N1;
 
    gy = linspace(0,1,N2);
    ny = N2;
    
    % grid the unitsquare
    %
    %  If you want to change the grid, do it here, for example
    %  if you want to have more nodes at the edges or gauss nodes
    %  or whatever
    %
    npoints = 1;
    for j1 = 1:ny
        for j2 = 1:nx
            z(:,npoints) = [gx(j2);gy(j1)];
            npoints = npoints+1;
        end
    end
    npoints = npoints - 1;
    
    for j = 1:npoints
        nodes(:,ncounter) = x(:,1)*(1-z(1,j))*(1-z(2,j));
        nodes(:,ncounter) = nodes(:,ncounter) + x(:,2)*z(1,j)*(1-z(2,j));
        nodes(:,ncounter) = nodes(:,ncounter) + x(:,3)*z(1,j)*z(2,j);
        nodes(:,ncounter) = nodes(:,ncounter) + x(:,4)*(1-z(1,j))*z(2,j);
        ncounter = ncounter + 1;
    end
    
    e1 = [1:nx-1]';
    for j = 1:ny-1
        elem(ecounter:ecounter+nx-2,1) = e1;
        elem(ecounter:ecounter+nx-2,2) = e1 + 1;
        elem(ecounter:ecounter+nx-2,3) = e1 + nx + 1;
        elem(ecounter:ecounter+nx-2,4) = e1 + nx;
        ecounter = ecounter + nx - 1;
        e1 = e1 + nx;
    end
    if (triangles)
        [n,m] = size(elem);
        %            % check the edge length of the rectangle
        %n1 = nodes(j,1);
        %n2 = nodes(j,2);
        %n3 = nodes(j,3);
        %if( norm(n2 - n1) < norm(n3 - n2) )
        elem3(2*[1:n] - 1,1) = elem([1:n],1);
        elem3(2*[1:n] - 1,2) = elem([1:n],2);
        elem3(2*[1:n] - 1,3) = elem([1:n],3);
        elem3(2*[1:n], 1) = elem([1:n],1);
        elem3(2*[1:n], 2) = elem([1:n],3);
        elem3(2*[1:n], 3) = elem([1:n],4);
        clear elem;
        elem = elem3;
    end
end



function n = myrows(A)
  [n,~] = size(A);
end

function n = mycolumns(A)
  [~,n] = size(A);
end
